# -*- coding: utf-8 -*-

"""Upload the Biomappings to NDEx.

.. seealso:: https://www.ndexbio.org/viewer/networks/402d1fd6-49d6-11eb-9e72-0ac135e8bacf
"""

import click
import pystow
from tqdm import tqdm

from biomappings import load_mappings
from biomappings.utils import MiriamValidator, get_git_hash

BIOMAPPINGS_NDEX_UUID = '402d1fd6-49d6-11eb-9e72-0ac135e8bacf'


@click.command()
@click.option('--username')
@click.option('--password')
def ndex(username, password):
    """Upload to NDEx."""
    try:
        from ndex2 import NiceCXBuilder
    except ImportError:
        click.secho('Need to `pip install ndex2` before uploading to NDEx', fg='red')
        return

    miriam_validator = MiriamValidator()
    positive_mappings = load_mappings()
    cx = NiceCXBuilder()
    cx.set_name('Biomappings')
    cx.add_network_attribute('description', 'Manually curated mappings (skos:exactMatch) between biological entities.')
    cx.add_network_attribute('reference', 'https://github.com/biomappings/biomappings')
    cx.add_network_attribute('rights', 'Waiver-No rights reserved (CC0)')

    context = {'orcid': 'https://identifiers.org/orcid:'}
    for mapping in positive_mappings:
        for prefix in (mapping['source prefix'], mapping['target prefix']):
            if miriam_validator.namespace_embedded(prefix):
                prefix = prefix.upper()
            context[prefix] = f'https://identifiers.org/{prefix}:'
    cx.set_context(context)

    cx.add_network_attribute('version', get_git_hash())
    authors = sorted(set(
        mapping['source']
        for mapping in positive_mappings
        if mapping['source'].startswith('orcid:')
    ))
    cx.add_network_attribute('author', authors, type='list_of_string')

    for mapping in tqdm(positive_mappings, desc='Loading NiceCXBuilder'):
        source = cx.add_node(
            represents=mapping['source name'],
            name=miriam_validator.get_curie(mapping["source prefix"], mapping["source identifier"]),
        )
        target = cx.add_node(
            represents=mapping['target name'],
            name=miriam_validator.get_curie(mapping["target prefix"], mapping["target identifier"]),
        )
        edge = cx.add_edge(
            source=source,
            target=target,
            interaction=mapping['relation'],
        )
        cx.add_edge_attribute(edge, 'type', mapping['type'])
        cx.add_edge_attribute(edge, 'provenance', mapping['source'])

    nice_cx = cx.get_nice_cx()
    nice_cx.update_to(
        uuid=BIOMAPPINGS_NDEX_UUID,
        server=pystow.get_config('ndex', 'server', 'http://public.ndexbio.org'),
        username=pystow.get_config('ndex', 'username', username),
        password=pystow.get_config('ndex', 'password', password),
    )


if __name__ == '__main__':
    ndex()



# -*- coding: utf-8 -*-

"""Utilities."""

import os
import re
from subprocess import CalledProcessError, check_output  # noqa: S404
from typing import Any, Mapping, Optional, Tuple

import requests

HERE = os.path.dirname(os.path.abspath(__file__))
RESOURCE_PATH = os.path.abspath(os.path.join(HERE, 'resources'))
IMG = os.path.abspath(os.path.join(HERE, os.pardir, os.pardir, 'docs', 'img'))


def get_git_hash() -> Optional[str]:
    """Get the git hash.

    :return:
        The git hash, equals 'UNHASHED' if encountered CalledProcessError, signifying that the
        code is not installed in development mode.
    """
    rv = _git('rev-parse', 'HEAD')
    if rv:
        return rv[:6]


def commit(message: str) -> Optional[str]:
    """Make a commit with the following message."""
    return _git('commit', '-m', message, '-a')


def push() -> Optional[str]:
    """Push the git repo."""
    return _git('push')


def not_main() -> bool:
    """Return if on the master branch."""
    return 'master' != _git('rev-parse', '--abbrev-ref', 'HEAD')


def _git(*args: str) -> Optional[str]:
    with open(os.devnull, 'w') as devnull:
        try:
            ret = check_output(  # noqa: S603,S607
                ['git', *args],
                cwd=os.path.dirname(__file__),
                stderr=devnull,
            )
        except CalledProcessError:
            return
        else:
            return ret.strip().decode('utf-8')


def get_script_url(fname: str) -> str:
    """Get the source path for this script.

    :param fname: Pass ``__file__`` as the argument to this function.
    :return: The script's URL to GitHub
    """
    commit_hash = get_git_hash()
    script_name = os.path.basename(fname)
    return f'https://github.com/biomappings/biomappings/blob/{commit_hash}/scripts/{script_name}'


def get_canonical_tuple(mapping: Mapping[str, Any]) -> Tuple[str, str, str, str]:
    """Get the canonical tuple from a mapping entry."""
    source = mapping['source prefix'], mapping['source identifier']
    target = mapping['target prefix'], mapping['target identifier']
    if source > target:
        source, target = target, source
    return (*source, *target)


class InvalidPrefix(ValueError):
    """Raised for an invalid prefix."""


class InvalidIdentifier(ValueError):
    """Raised for an invalid identifier."""


class MiriamValidator:
    """Validate prefix/identifier pairs based on the MIRIAM database."""

    def __init__(self):  # noqa: D107
        self.entries = self._load_identifiers_entries()

    @staticmethod
    def _load_identifiers_entries():
        url = 'https://registry.api.identifiers.org/resolutionApi/getResolverDataset'
        res = requests.get(url)
        regj = res.json()
        patterns = {
            entry['prefix']: {
                'pattern': re.compile(entry['pattern']),
                'namespace_embedded': entry['namespaceEmbeddedInLui'],
            }
            for entry in sorted(regj['payload']['namespaces'], key=lambda x: x['prefix'])
        }
        return patterns

    def namespace_embedded(self, prefix: str) -> bool:
        """Return True if the namespace is embedded for the given prefix."""
        return self.entries[prefix]['namespace_embedded']

    def check_valid_prefix_id(self, prefix, identifier):
        """Check the prefix/identifier pair is valid."""
        if prefix not in self.entries:
            raise InvalidPrefix(prefix)
        entry = self.entries[prefix]
        if not re.match(entry['pattern'], identifier):
            raise InvalidIdentifier(identifier)

    def get_curie(self, prefix: str, identifier: str) -> str:
        """Return CURIE for a given prefix and identifier."""
        if self.namespace_embedded(prefix):
            return identifier
        else:
            return f'{prefix}:{identifier}'

    def get_url(self, prefix: str, identifier: str) -> str:
        """Return URL for a given prefix and identifier."""
        return f'https://identifiers.org/{self.get_curie(prefix, identifier)}'



# -*- coding: utf-8 -*-

"""Generate a summary for the Biomappings website."""

import itertools as itt
import os
from collections import Counter
from typing import Iterable, Mapping, Optional

import click
import yaml

__all__ = [
    'export',
]


@click.command()
def export():
    """Create export data file."""
    from biomappings.resources import load_mappings, load_predictions, load_false_mappings

    here = os.path.abspath(os.path.dirname(__file__))
    path = os.path.join(here, os.pardir, os.pardir, 'docs', '_data', 'summary.yml')

    true_mappings = load_mappings()
    false_mappings = load_false_mappings()
    rv = {
        'positive': _get_counter(true_mappings),
        'negative': _get_counter(false_mappings),
        'predictions': _get_counter(load_predictions()),
        'contributors': _get_contributors(itt.chain(true_mappings, false_mappings)),
    }
    rv.update({
        f'{k}_mapping_count': sum(e['count'] for e in rv[k])
        for k in ('positive', 'negative', 'predictions')
    })
    rv.update({
        f'{k}_prefix_count': len(set(itt.chain.from_iterable((e['source'], e['target']) for e in rv[k])))
        for k in ('positive', 'negative', 'predictions')
    })
    with open(path, 'w') as file:
        yaml.safe_dump(rv, file, indent=2)


def _get_counter(mappings: Iterable[Mapping[str, str]]):
    counter = Counter()
    for mapping in mappings:
        source, target = mapping['source prefix'], mapping['target prefix']
        if source > target:
            source, target = target, source
        counter[source, target] += 1
    return [
        dict(source=source, target=target, count=count)
        for (source, target), count in counter.most_common()
    ]


def _get_contributors(mappings: Iterable[Mapping[str, str]]):
    counter = Counter(
        _get_source(mapping['source'])
        for mapping in mappings
    )
    return [
        dict(orcid=orcid, count=count) if orcid else dict(count=count)
        for orcid, count in counter.most_common()
    ]


def _get_source(source: str) -> Optional[str]:
    if source.startswith('orcid:'):
        return source[len('orcid:'):]


if __name__ == '__main__':
    export()



"""Validation tests for :mod:`biomappings`."""

import itertools as itt
from collections import Counter

from biomappings import load_false_mappings, load_mappings, load_predictions
from biomappings.utils import MiriamValidator, get_canonical_tuple

mappings = load_mappings()
predictions = load_predictions()
incorrect = load_false_mappings()


miriam_validator = MiriamValidator()


def test_valid_mappings():
    """Test the validity of the prefixes and identifiers in the mappings."""
    for mapping in itt.chain(mappings, incorrect, predictions):
        miriam_validator.check_valid_prefix_id(
            mapping['source prefix'],
            mapping['source identifier'],
        )
        miriam_validator.check_valid_prefix_id(
            mapping['target prefix'],
            mapping['target identifier'],
        )


def test_redundancy():
    """Test the redundancy of manually curated mappings and predicted mappings."""
    counter = Counter(
        get_canonical_tuple(m)
        for m in itt.chain(mappings, incorrect, predictions)
    )
    redundant = [(k, v) for k, v in counter.items() if v > 1]
    if redundant:
        r = '\n'.join(f'  {r}: {count}' for r, count in redundant)
        raise ValueError(f'Redundant: {r}')



# -*- coding: utf-8 -*-

"""The biomappings CLI."""

import click

from .graph import charts
from .summary import export
from .upload_ndex import ndex


@click.group()
def main():
    """Run the biomappings CLI."""


@main.command()
def web():
    """Run the biomappings web curation interface."""
    from .wsgi import app
    from .utils import get_git_hash
    if get_git_hash() is None:
        click.secho(
            'You are not running biomappings from a development installation.\n'
            'Please run the following to install in development mode:\n'
            '  $ git clone https://github.com/biomappings/biomappings.git\n'
            '  $ cd biomappings\n'
            '  $ pip install -e .',
            fg='red',
        )
    else:
        app.run()


main.add_command(export)
main.add_command(ndex)
main.add_command(charts)

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

"""Load Biomappings as a graph."""

import os

import click
import matplotlib.pyplot as plt
import networkx as nx
import seaborn as sns

from biomappings.resources import load_mappings
from biomappings.utils import IMG, MiriamValidator


def get_positive_graph() -> nx.Graph:
    """Get a graph of the positive mappings."""
    v = MiriamValidator()
    graph = nx.Graph()
    for mapping in load_mappings():
        source_curie = v.get_curie(mapping['source prefix'], mapping['source identifier'])
        graph.add_node(
            source_curie,
            prefix=mapping['source prefix'],
            identifier=mapping['source identifier'],
            name=mapping['source name'],
        )
        target_curie = v.get_curie(mapping['target prefix'], mapping['target identifier'])
        graph.add_node(
            target_curie,
            prefix=mapping['target prefix'],
            identifier=mapping['target identifier'],
            name=mapping['target name'],
        )
        graph.add_edge(
            source_curie,
            target_curie,
            provenance=mapping['source'],
            type=mapping['type'],
        )
    return graph


@click.command()
def charts():
    """Make charts."""
    graph = get_positive_graph()
    node_sizes, edge_sizes, densities, n_prefixes = [], [], [], []
    n_duplicates = []
    for c in nx.connected_components(graph):
        node_sizes.append(len(c))
        sg = graph.subgraph(c)
        edge_sizes.append(sg.number_of_edges())
        densities.append(nx.density(sg))
        prefixes = [
            graph.nodes[node]['prefix']
            for node in c
        ]
        unique_prefixes = len(set(prefixes))
        n_prefixes.append(unique_prefixes)
        n_duplicates.append(len(prefixes) - unique_prefixes)

    fig, axes = plt.subplots(2, 3, figsize=(10, 6.5))

    sns.histplot(node_sizes, ax=axes[0][0])
    axes[0][0].set_yscale('log')
    axes[0][0].set_title('Component Node Sizes')

    sns.histplot(edge_sizes, ax=axes[0][1])
    axes[0][1].set_yscale('log')
    axes[0][1].set_title('Component Edge Sizes')
    axes[0][1].set_ylabel('')

    sns.kdeplot(densities, ax=axes[0][2], log_scale=True)
    # axes[2].set_xlim([0.0, 1.0])
    axes[0][2].set_title('Component Densities')
    axes[0][2].set_ylabel('')

    sns.histplot(n_prefixes, ax=axes[1][0])
    axes[1][0].set_title('Number Prefixes')

    # has duplicate prefix in component

    sns.histplot(n_duplicates, ax=axes[1][1])
    axes[1][1].set_yscale('log')
    axes[0][2].set_ylabel('')
    axes[1][1].set_title('Number Duplicate Prefixes')

    axes[1][2].axis('off')

    path = os.path.join(IMG, 'components.png')
    print('saving to', path)
    plt.tight_layout()
    plt.savefig(path, dpi=300)
    plt.close(fig)


if __name__ == '__main__':
    charts()



# -*- coding: utf-8 -*-

"""Biomappings resources."""

import csv
import itertools as itt
import os
from typing import Dict, Iterable, List, Mapping, Sequence, Tuple

from biomappings.utils import RESOURCE_PATH, get_canonical_tuple

MAPPINGS_HEADER = [
    'source prefix',
    'source identifier',
    'source name',
    'relation',
    'target prefix',
    'target identifier',
    'target name',
    'type',
    'source',
]
PREDICTIONS_HEADER = [
    'source prefix',
    'source identifier',
    'source name',
    'relation',
    'target prefix',
    'target identifier',
    'target name',
    'type',
    'confidence',
    'source',
]


def get_resource_file_path(fname) -> str:
    """Get a resource by its file name."""
    return os.path.join(RESOURCE_PATH, fname)


def _load_table(fname) -> List[Dict[str, str]]:
    with open(fname, 'r') as fh:
        reader = csv.reader(fh, delimiter='\t')
        header = next(reader)
        return [dict(zip(header, row)) for row in reader]


def _write_helper(header: Sequence[str], lod: Iterable[Mapping[str, str]], path: str, mode: str) -> None:
    with open(path, mode) as file:
        if mode == 'w':
            print(*header, sep='\t', file=file)
        for line in lod:
            print(*[line[k] for k in header], sep='\t', file=file)


def load_mappings() -> List[Dict[str, str]]:
    """Load the mappings table."""
    return _load_table(get_resource_file_path('mappings.tsv'))


def append_true_mappings(m: Iterable[Mapping[str, str]]) -> None:
    """Append new lines to the mappings table."""
    _write_helper(MAPPINGS_HEADER, m, get_resource_file_path('mappings.tsv'), 'a')


def load_false_mappings() -> List[Dict[str, str]]:
    """Load the false mappings table."""
    return _load_table(get_resource_file_path('incorrect.tsv'))


def append_false_mappings(m: Iterable[Mapping[str, str]]) -> None:
    """Append new lines to the false mappings table."""
    _write_helper(MAPPINGS_HEADER, m, get_resource_file_path('incorrect.tsv'), 'a')


def load_predictions() -> List[Dict[str, str]]:
    """Load the predictions table."""
    return _load_table(get_resource_file_path('predictions.tsv'))


def write_predictions(m: List[Mapping[str, str]]) -> None:
    """Write new content to the predictions table."""
    _write_helper(PREDICTIONS_HEADER, m, get_resource_file_path('predictions.tsv'), 'w')


def append_prediction_tuples(m: Iterable[Tuple[str, ...]], deduplicate: bool = True) -> None:
    """Append new lines to the predictions table that come as canonical tuples."""
    append_predictions(
        (
            dict(zip(PREDICTIONS_HEADER, p))
            for p in m
        ),
        deduplicate=deduplicate,
    )


def append_predictions(mappings: Iterable[Mapping[str, str]], deduplicate: bool = True) -> None:
    """Append new lines to the predictions table."""
    if deduplicate:
        existing_mappings = {
            get_canonical_tuple(existing_mapping)
            for existing_mapping in itt.chain(
                load_mappings(),
                load_false_mappings(),
                load_predictions(),
            )
        }
        mappings = (
            mapping
            for mapping in mappings
            if get_canonical_tuple(mapping) not in existing_mappings
        )

    _write_helper(PREDICTIONS_HEADER, mappings, get_resource_file_path('predictions.tsv'), 'a')



# -*- coding: utf-8 -*-

"""Web curation interface for :mod:`biomappings`."""

import getpass
import os
from typing import Any, Iterable, Mapping, Optional, Tuple

import flask
import flask_bootstrap
from flask_wtf import FlaskForm
from wtforms import StringField, SubmitField

from biomappings.resources import append_false_mappings, append_true_mappings, load_predictions, write_predictions
from biomappings.utils import MiriamValidator, commit, not_main, push

app = flask.Flask(__name__)
app.config['WTF_CSRF_ENABLED'] = False
app.config['SECRET_KEY'] = os.urandom(8)
app.config['SHOW_RELATIONS'] = True
flask_bootstrap.Bootstrap(app)

# A mapping from your computer's user, returned by getuser.getpass()
KNOWN_USERS = {
    'cthoyt': '0000-0003-4423-4370',
    'ben': '0000-0001-9439-5346',
}


def _manual_source():
    known_user = KNOWN_USERS.get(getpass.getuser())
    if known_user:
        return f'orcid:{known_user}'
    return 'web'


miriam_validator = MiriamValidator()


class Controller:
    """A module for interacting with the predictions and mappings."""

    def __init__(self):  # noqa: D107
        self._predictions = load_predictions()
        self._marked = {}
        self.total_curated = 0
        self._added_mappings = []

    def predictions(
        self,
        *,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        query: Optional[str] = None,
        prefix: Optional[str] = None,
    ) -> Iterable[Tuple[int, Mapping[str, Any]]]:
        """Iterate over predictions.

        :param offset: If given, offset the iteration by this number
        :param limit: If given, only iterate this number of predictions.
        :param query: If given, show only equivalences that have it appearing as a substring in one of the fields
        :param prefix: If given, show only equivalences that have it appearing as a substring in one of the prefixes
        :yields: Pairs of positions and prediction dictionaries
        """
        it = enumerate(self._predictions)
        if query is not None:
            query = query.casefold()
            it = (
                (line, prediction)
                for line, prediction in it
                if any(
                    query in prediction[x].casefold()
                    for x in ('source identifier', 'source name', 'target identifier', 'target name')
                )
            )
        if prefix is not None:
            prefix = prefix.casefold()
            it = (
                (line, prediction)
                for line, prediction in it
                if any(
                    prefix in prediction[x].casefold()
                    for x in ('source prefix', 'target prefix')
                )
            )

        it = (
            (line, prediction)
            for line, prediction in it
            if line not in self._marked
        )
        if offset is not None:
            try:
                for _ in range(offset):
                    next(it)
            except StopIteration:
                # if next() fails, then there are no remaining entries.
                # do not pass go, do not collect 200 euro $
                return
        if limit is None:
            yield from it
        else:
            for line_prediction, _ in zip(it, range(limit)):
                yield line_prediction

    @staticmethod
    def get_curie(prefix: str, identifier: str) -> str:
        """Return CURIE for a given prefix and identifier."""
        return miriam_validator.get_curie(prefix, identifier)

    @classmethod
    def get_url(cls, prefix: str, identifier: str) -> str:
        """Return URL for a given prefix and identifier."""
        return miriam_validator.get_url(prefix, identifier)

    @property
    def total_predictions(self) -> int:
        """Return the total number of yet unmarked predictions."""
        return len(self._predictions) - len(self._marked)

    def mark(self, line: int, correct: bool) -> None:
        """Mark the given equivalency as correct.

        :param line: Position of the prediction
        :param correct: Value to mark the prediction with
        """
        if line not in self._marked:
            self.total_curated += 1
        self._marked[line] = correct

    def add_mapping(
        self,
        source_prefix: str,
        source_id: str,
        source_name: str,
        target_prefix: str,
        target_id: str,
        target_name: str,
    ) -> None:
        """Add manually curated new mappings."""
        try:
            miriam_validator.check_valid_prefix_id(source_prefix, source_id)
        except ValueError as e:
            flask.flash(
                f'Problem with source CURIE {source_prefix}:{source_id}: {e.__class__.__name__}',
                category='warning',
            )
            return

        try:
            miriam_validator.check_valid_prefix_id(target_prefix, target_id)
        except ValueError as e:
            flask.flash(
                f'Problem with target CURIE {target_prefix}:{target_id}: {e.__class__.__name__}',
                category='warning',
            )
            return

        self._added_mappings.append({
            'source prefix': source_prefix,
            'source identifier': source_id,
            'source name': source_name,
            'relation': 'skos:exactMatch',
            'target prefix': target_prefix,
            'target identifier': target_id,
            'target name': target_name,
            'source': _manual_source(),
            'type': 'manual',
        })
        self.total_curated += 1

    def persist(self):
        """Save the current markings to the source files."""
        curated_true_entries = []
        curated_false_entries = []

        for line, correct in sorted(self._marked.items(), reverse=True):
            prediction = self._predictions.pop(line)
            prediction['source'] = _manual_source()
            prediction['type'] = 'manually_reviewed'
            if correct:
                curated_true_entries.append(prediction)
            else:
                curated_false_entries.append(prediction)

        append_true_mappings(curated_true_entries)
        append_false_mappings(curated_false_entries)
        write_predictions(self._predictions)
        self._marked.clear()

        # Now add manually curated mappings
        append_true_mappings(self._added_mappings)
        self._added_mappings = []


controller = Controller()


class MappingForm(FlaskForm):
    """Form for entering new mappings."""

    source_prefix = StringField('Source Prefix', id='source_prefix')
    source_id = StringField('Source ID', id='source_id')
    source_name = StringField('Source Name', id='source_name')
    target_prefix = StringField('Target Prefix', id='target_prefix')
    target_id = StringField('Target ID', id='target_id')
    target_name = StringField('Target Name', id='target_name')
    submit = SubmitField('Add')


@app.route('/')
def home():
    """Serve the home page."""
    form = MappingForm()
    limit = flask.request.args.get('limit', type=int, default=10)
    offset = flask.request.args.get('offset', type=int, default=0)
    query = flask.request.args.get('query')
    prefix = flask.request.args.get('prefix')
    show_relations = app.config['SHOW_RELATIONS']
    return flask.render_template(
        'home.html',
        controller=controller,
        form=form,
        limit=limit,
        offset=offset,
        query=query,
        prefix=prefix,
        show_relations=show_relations,
    )


@app.route('/add_mapping', methods=['POST'])
def add_mapping():
    """Add a new mapping manually."""
    form = MappingForm()
    if form.is_submitted():
        controller.add_mapping(
            form.data['source_prefix'], form.data['source_id'], form.data['source_name'],
            form.data['target_prefix'], form.data['target_id'], form.data['target_name'],
        )
        controller.persist()
    else:
        flask.flash('missing form data', category='warning')
    return _go_home()


@app.route('/commit')
def run_commit():
    """Make a commit then redirect to the the home page."""
    commit(
        f'Curated {controller.total_curated} mapping{"s" if controller.total_curated > 1 else ""}'
        f' ({getpass.getuser()})',
    )
    if not_main():
        push()
    controller.total_curated = 0
    return _go_home()


@app.route('/mark/<int:line>/<value>')
def mark(line: int, value: str):
    """Mark the given line as correct or not."""
    controller.mark(line, value.lower() in {'yup', 'true', 't', 'correct', 'right', 'close enough', 'disco'})
    controller.persist()
    return _go_home()


def _go_home():
    return flask.redirect(flask.url_for(
        'home',
        limit=flask.request.args.get('limit', type=int),
        offset=flask.request.args.get('offset', type=int),
        query=flask.request.args.get('query'),
        prefix=flask.request.args.get('prefix'),
        show_relations=app.config['SHOW_RELATIONS'],
    ))


if __name__ == '__main__':
    app.run()
